from typing import *
import royalnet
import royalnet.commands as rc
import royalnet.utils as ru
from ..tables.telegram import Telegram
from ..tables.discord import Discord


class RoyalnetsyncCommand(rc.Command):
    name: str = "royalnetsync"

    description: str = "Connect your chat account to Royalnet!"

    syntax: str = "{username} {password}"

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        username = args[0]
        password = " ".join(args[1:])

        author = await data.get_author(error_if_none=True)

        user = await data.find_user(username)
        try:
            successful = user.test_password(password)
        except ValueError:
            raise rc.UserError(f"User {user} has no password set!")
        if not successful:
            raise rc.InvalidInputError(f"Invalid password!")

        if self.interface.name == "telegram":
            import telegram
            message: telegram.Message = data.message
            from_user: telegram.User = message.from_user
            TelegramT = self.alchemy.get(Telegram)
            tg_user: Telegram = await ru.asyncify(
                data.session.query(TelegramT).filter_by(tg_id=from_user.id).one_or_none
            )
            if tg_user is None:
                # Create
                tg_user = TelegramT(
                    user=author,
                    tg_id=from_user.id,
                    first_name=from_user.first_name,
                    last_name=from_user.last_name,
                    username=from_user.username
                )
                data.session.add(tg_user)
            else:
                # Edit
                tg_user.first_name = from_user.first_name
                tg_user.last_name = from_user.last_name
                tg_user.username = from_user.username
            await data.session_commit()
            await data.reply(f"↔️ Account {tg_user} synced to {author}!")

        elif self.interface.name == "discord":
            import discord
            message: discord.Message = data.message
            author: discord.User = message.author
            DiscordT = self.alchemy.get(Discord)
            ds_user: Discord = await ru.asyncify(
                data.session.query(DiscordT).filter_by(discord_id=author.id).one_or_none
            )
            if ds_user is None:
                # Create
                ds_user = DiscordT(
                    user=author,
                    discord_id=author.id,
                    username=author.name,
                    discriminator=author.discriminator,
                    avatar_url=author.avatar_url
                )
                data.session.add(ds_user)
            else:
                # Edit
                ds_user.username = author.name
                ds_user.discriminator = author.discriminator
                ds_user.avatar_url = author.avatar_url
            await data.session_commit()
            await data.reply(f"↔️ Account {ds_user} synced to {author}!")

        elif self.interface.name == "matrix":
            raise rc.UnsupportedError(f"{self} hasn't been implemented for Matrix yet")

        else:
            raise rc.UnsupportedError(f"Unknown interface: {self.interface.name}")



def escape(string: str) -> str:
    """Escape a string to be sent through Matrix, and format it using RoyalCode.

    Underlines are currently unsupported.

    Warning:
        Currently escapes everything, even items in code blocks."""
    return string.replace("[b]", "**") \
        .replace("[/b]", "**") \
        .replace("[i]", "_") \
        .replace("[/i]", "_") \
        .replace("[c]", "`") \
        .replace("[/c]", "`") \
        .replace("[p]", "```") \
        .replace("[/p]", "```")



import click
import multiprocessing
import toml
import logging
import royalnet.utils as ru

try:
    import royalnet.serf.telegram as rst
except ImportError:
    rst = None

try:
    import royalnet.serf.discord as rsd
except ImportError:
    rsd = None

try:
    import royalnet.serf.matrix as rsm
except ImportError:
    rsm = None

try:
    import royalnet.constellation as rc
except ImportError:
    rc = None

try:
    import royalnet.herald as rh
except ImportError:
    rh = None

try:
    import coloredlogs
except ImportError:
    coloredlogs = None


log = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--config-filename", default="./config.toml", type=click.Path(exists=True),
              help="The filename of the Royalnet configuration file.")
def run(config_filename: str):
    # Read the configuration file
    with open(config_filename, "r") as t:
        config: dict = toml.load(t)

    ru.init_logging(config["Logging"])

    if config["Sentry"] is None or not config["Sentry"]["enabled"]:
        log.info("Sentry: disabled")
    else:
        try:
            ru.init_sentry(config["Sentry"])
        except ImportError:
            log.info("Sentry: not installed")

    # Herald Server
    herald_cfg = None
    herald_process = None
    if rh is not None and "Herald" in config:
        if "Local" in config["Herald"] and config["Herald"]["Local"]["enabled"]:
            # Create a Herald server
            herald_server = rh.Server(rh.Config.from_config(name="<server>", **config["Herald"]["Local"]))
            # Run the Herald server on a new process
            herald_process = multiprocessing.Process(name="Herald.Local",
                                                     target=herald_server.run_blocking,
                                                     daemon=True,
                                                     kwargs={
                                                         "logging_cfg": config["Logging"]
                                                     })
            herald_process.start()
            herald_cfg = config["Herald"]["Local"]
            log.info("Herald: Enabled (Local)")
        elif "Remote" in config["Herald"] and config["Herald"]["Remote"]["enabled"]:
            log.info("Herald: Enabled (Remote)")
            herald_cfg = config["Herald"]["Remote"]
        else:
            log.info("Herald: Disabled")
    else:
        log.info("Herald: Disabled")

    # Serfs
    telegram_process = None
    if rst is not None and "Telegram" in config["Serfs"] and config["Serfs"]["Telegram"]["enabled"]:
        telegram_process = multiprocessing.Process(name="Serf.Telegram",
                                                   target=rst.TelegramSerf.run_process,
                                                   daemon=True,
                                                   kwargs={
                                                       "alchemy_cfg": config["Alchemy"],
                                                       "herald_cfg": herald_cfg,
                                                       "packs_cfg": config["Packs"],
                                                       "sentry_cfg": config["Sentry"],
                                                       "logging_cfg": config["Logging"],
                                                       "serf_cfg": config["Serfs"]["Telegram"],
                                                   })
        telegram_process.start()
        log.info("Serf.Telegram: Started")
    else:
        log.info("Serf.Telegram: Disabled")

    discord_process = None
    if rsd is not None and "Discord" in config["Serfs"] and config["Serfs"]["Discord"]["enabled"]:
        discord_process = multiprocessing.Process(name="Serf.Discord",
                                                  target=rsd.DiscordSerf.run_process,
                                                  daemon=True,
                                                  kwargs={
                                                      "alchemy_cfg": config["Alchemy"],
                                                      "herald_cfg": herald_cfg,
                                                      "packs_cfg": config["Packs"],
                                                      "sentry_cfg": config["Sentry"],
                                                      "logging_cfg": config["Logging"],
                                                      "serf_cfg": config["Serfs"]["Discord"],
                                                  })
        discord_process.start()
        log.info("Serf.Discord: Started")
    else:
        log.info("Serf.Discord: Disabled")

    matrix_process = None
    if rsm is not None and "Matrix" in config["Serfs"] and config["Serfs"]["Matrix"]["enabled"]:
        matrix_process = multiprocessing.Process(name="Serf.Matrix",
                                                 target=rsm.MatrixSerf.run_process,
                                                 daemon=True,
                                                 kwargs={
                                                     "alchemy_cfg": config["Alchemy"],
                                                     "herald_cfg": herald_cfg,
                                                     "packs_cfg": config["Packs"],
                                                     "sentry_cfg": config["Sentry"],
                                                     "logging_cfg": config["Logging"],
                                                     "serf_cfg": config["Serfs"]["Matrix"],
                                                 })
        matrix_process.start()
        log.info("Serf.Matrix: Started")
    else:
        log.info("Serf.Matrix: Disabled")

    # Constellation
    constellation_process = None
    if rc is not None and "Constellation" in config and config["Constellation"]["enabled"]:
        constellation_process = multiprocessing.Process(name="Constellation",
                                                        target=rc.Constellation.run_process,
                                                        daemon=True,
                                                        kwargs={
                                                            "alchemy_cfg": config["Alchemy"],
                                                            "herald_cfg": herald_cfg,
                                                            "packs_cfg": config["Packs"],
                                                            "sentry_cfg": config["Sentry"],
                                                            "logging_cfg": config["Logging"],
                                                            "constellation_cfg": config["Constellation"],
                                                        })
        constellation_process.start()
        log.info("Constellation: Started")
    else:
        log.info("Constellation: Disabled")

    log.info("All processes started!")
    if constellation_process is not None:
        log.info("Waiting for Constellation to stop...")
        constellation_process.join()
    if telegram_process is not None:
        log.info("Waiting for Serf.Telegram to stop...")
        telegram_process.join()
    if discord_process is not None:
        log.info("Waiting for Serf.Discord to stop...")
        discord_process.join()
    if matrix_process is not None:
        log.info("Waiting for Serf.Matrix to stop...")
        matrix_process.join()
    if herald_process is not None:
        log.info("Waiting for Herald to stop...")
        herald_process.join()


if __name__ == "__main__":
    run()



# Imports go here!
from .royalnetversion import RoyalnetversionCommand
from .royalnetsync import RoyalnetsyncCommand
from .royalnetroles import RoyalnetrolesCommand
from .royalnetaliases import RoyalnetaliasesCommand

# Enter the commands of your Pack here!
available_commands = [
    RoyalnetversionCommand,
    RoyalnetsyncCommand,
    RoyalnetrolesCommand,
    RoyalnetaliasesCommand,
]

# Don't change this, it should automatically generate __all__
__all__ = [command.__name__ for command in available_commands]



# Imports go here!
from .exception import ExceptionEvent

# Enter the commands of your Pack here!
available_events = [
    ExceptionEvent,
]

# Don't change this, it should automatically generate __all__
__all__ = [command.__name__ for command in available_events]



import typing
import re
import os
import logging
import ffmpeg
import discord
from contextlib import asynccontextmanager
from royalnet.utils import asyncify, MultiLock
from royalnet.bard import YtdlInfo, YtdlFile
from .fileaudiosource import FileAudioSource

log = logging.getLogger(__name__)


class YtdlDiscord:
    """A representation of a :class:`YtdlFile` conversion to the :mod:`discord` PCM format."""

    def __init__(self, ytdl_file: YtdlFile):
        self.ytdl_file: YtdlFile = ytdl_file
        self.pcm_filename: typing.Optional[str] = None
        self.lock: MultiLock = MultiLock()

    def __repr__(self):
        if not self.ytdl_file.has_info:
            return f"<{self.__class__.__qualname__} without info>"
        elif not self.ytdl_file.is_downloaded:
            return f"<{self.__class__.__qualname__} not downloaded>"
        elif not self.is_converted:
            return f"<{self.__class__.__qualname__} at '{self.ytdl_file.filename}' not converted>"
        else:
            return f"<{self.__class__.__qualname__} at '{self.pcm_filename}'>"

    @property
    def is_converted(self):
        """Has the file been converted?"""
        return self.pcm_filename is not None

    async def convert_to_pcm(self) -> None:
        """Convert the file to pcm with :mod:`ffmpeg`."""
        await self.ytdl_file.download_file()
        if self.pcm_filename is None:
            async with self.ytdl_file.lock.normal():
                destination_filename = re.sub(r"\.[^.]+$", ".pcm", self.ytdl_file.filename)
                async with self.lock.exclusive():
                    log.debug(f"Converting to PCM: {self.ytdl_file.filename}")
                    out, err = await asyncify(
                        ffmpeg.input(self.ytdl_file.filename)
                              .output(destination_filename, format="s16le", ac=2, ar="48000")
                              .overwrite_output()
                              .run,
                        capture_stdout=True,
                        capture_stderr=True,
                    )
                    log.debug(f"ffmpeg returned: ({type(out)}, {type(err)})")
            self.pcm_filename = destination_filename

    async def delete_asap(self) -> None:
        """Delete the mp3 file."""
        log.debug(f"Trying to delete: {self}")
        if self.is_converted:
            async with self.lock.exclusive():
                os.remove(self.pcm_filename)
                log.debug(f"Deleted: {self.pcm_filename}")
                self.pcm_filename = None

    @classmethod
    async def from_url(cls, url, **ytdl_args) -> typing.List["YtdlDiscord"]:
        """Create a :class:`list` of :class:`YtdlMp3` from a URL."""
        files = await YtdlFile.from_url(url, **ytdl_args)
        dfiles = []
        for file in files:
            dfile = YtdlDiscord(file)
            dfiles.append(dfile)
        return dfiles

    @property
    def info(self) -> typing.Optional[YtdlInfo]:
        """Shortcut to get the :class:`YtdlInfo` of the object."""
        return self.ytdl_file.info

    @asynccontextmanager
    async def spawn_audiosource(self):
        log.debug(f"Spawning audio_source for: {self}")
        await self.convert_to_pcm()
        async with self.lock.normal():
            with open(self.pcm_filename, "rb") as stream:
                fas = FileAudioSource(stream)
                yield fas

    def embed(self) -> "discord.Embed":
        """Return this info as a :py:class:`discord.Embed`."""
        colors = {
            "youtube": 0xCC0000,
            "soundcloud": 0xFF5400,
            "Clyp": 0x3DBEB3,
            "Bandcamp": 0x1DA0C3,
            "PeerTube": 0xF1680D,
            "generic": 0x4F545C,
        }
        embed = discord.Embed(title=self.info.title,
                              colour=discord.Colour(colors.get(self.info.extractor, 0x4F545C)),
                              url=self.info.webpage_url if (self.info.webpage_url and self.info.webpage_url.startswith("http")) else discord.embeds.EmptyEmbed)
        if self.info.thumbnail:
            embed.set_thumbnail(url=self.info.thumbnail)
        if self.info.uploader:
            embed.set_author(name=self.info.uploader,
                             url=self.info.uploader_url if self.info.uploader_url is not None else discord.embeds.EmptyEmbed)
        elif self.info.artist:
            embed.set_author(name=self.info.artist,
                             url=discord.embeds.EmptyEmbed)
        if self.info.album:
            embed.add_field(name="Album", value=self.info.album, inline=True)
        if self.info.duration:
            embed.add_field(name="Duration", value=str(self.info.duration), inline=True)
        if self.info.extractor != "generic" and self.info.upload_date:
            embed.add_field(name="Published on", value=self.info.upload_date.strftime("%d %b %Y"), inline=True)
        # embed.set_footer(text="Source: youtube-dl", icon_url="https://i.imgur.com/TSvSRYn.png")
        return embed



from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       BigInteger, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .users import User


class Telegram:
    __tablename__ = "telegram"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"))

    @declared_attr
    def user(self):
        return relationship("User", backref="telegram")

    @declared_attr
    def tg_id(self):
        return Column(BigInteger, primary_key=True)

    @declared_attr
    def first_name(self):
        return Column(String)

    @declared_attr
    def last_name(self):
        return Column(String)

    @declared_attr
    def username(self):
        return Column(String)

    def __repr__(self):
        return f"<Telegram {str(self)}>"

    def __str__(self):
        return f"[c]telegram:{self.mention()}[/c]"

    def mention(self) -> str:
        if self.username is not None:
            return f"@{self.username}"
        elif self.last_name is not None:
            return f"{self.first_name} {self.last_name}"
        else:
            return self.first_name



from sqlalchemy.inspection import inspect
from sqlalchemy.schema import Table


def table_dfs(starting_table: Table, ending_table: Table) -> tuple:
    """Depth-first-search for the path from the starting table to the ending table.

    Returns:
        A :class:`tuple` containing the path, starting from the starting table and ending at the ending table."""
    inspected = set()

    def search(_mapper, chain):
        inspected.add(_mapper)
        if _mapper.class_ == ending_table:
            return chain
        relationships = _mapper.relationships
        for _relationship in set(relationships):
            if _relationship.mapper in inspected:
                continue
            result = search(_relationship.mapper, chain + (_relationship,))
            if len(result) != 0:
                return result
        return ()

    return search(inspect(starting_table), tuple())



from typing import *
import logging
import datetime
import asyncio as aio
import royalnet.backpack as rb
import royalnet.commands as rc
import royalnet.utils as ru
from ..serf import Serf
from .escape import escape


try:
    import nio
except ImportError:
    nio = None


log = logging.getLogger(__name__)


class MatrixSerf(Serf):
    """A serf that connects to `Matrix <https://matrix.org/>`_ as an user."""
    interface_name = "matrix"

    _identity_table = rb.tables.Matrix
    _identity_column = "matrix_id"

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: Dict[str, Any],
                 herald_cfg: Dict[str, Any],
                 sentry_cfg: Dict[str, Any],
                 packs_cfg: Dict[str, Any],
                 serf_cfg: Dict[str, Any],
                 **_):
        if nio is None:
            raise ImportError("'matrix' extra is not installed")

        super().__init__(loop=loop,
                         alchemy_cfg=alchemy_cfg,
                         herald_cfg=herald_cfg,
                         sentry_cfg=sentry_cfg,
                         packs_cfg=packs_cfg,
                         serf_cfg=serf_cfg)

        self.client: Optional[nio.AsyncClient] = None

        self.homeserver: str = serf_cfg["homeserver"]
        self.matrix_id: str = serf_cfg["matrix_id"]
        self.password: str = serf_cfg["password"]

        self._started_timestamp: Optional[int] = None

        self.Data: Type[rc.CommandData] = self.data_factory()

    def interface_factory(self) -> Type[rc.CommandInterface]:
        # noinspection PyPep8Naming
        GenericInterface = super().interface_factory()

        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordInterface(GenericInterface):
            name = self.interface_name
            prefix = "!"

        return DiscordInterface

    def data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters,PyAbstractClass
        class MatrixData(rc.CommandData):
            def __init__(data,
                         interface: rc.CommandInterface,
                         loop: aio.AbstractEventLoop,
                         room: nio.MatrixRoom,
                         event: nio.Event):
                super().__init__(interface=interface, loop=loop)
                data.room: nio.MatrixRoom = room
                data.event: nio.Event = event

            async def reply(data, text: str):
                await self.client.room_send(room_id=data.room.room_id, message_type="m.room.message", content={
                    "msgtype": "m.text",
                    "body": escape(text)
                })

            async def get_author(data, error_if_none=False):
                user: str = data.event.sender
                query = data.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user)
                result = await ru.asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise rc.CommandError("You must be registered to use this command.")
                return result

            # Delete invoking does not really make sense on Matrix

        return MatrixData

    async def handle_message(self, room: "nio.MatrixRoom", event: "nio.RoomMessageText"):
        # Skip events happened before the startup of the Serf
        if event.server_timestamp < self._started_timestamp:
            return
        # Find the text in the event
        text = event.body
        # Skip non-command events
        if not text.startswith("!"):
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        # Don't use a case-sensitive command name
        command_name = command_text.lower()
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            return
        # Send typing
        await self.client.room_typing(room_id=room.room_id, typing_state=True)
        # Open an alchemy session, if available
        if self.alchemy is not None:
            session = await ru.asyncify(self.alchemy.Session)
        else:
            session = None
        # Prepare data
        data = self.Data(interface=command.interface, loop=self.loop, room=room, event=event)
        # Call the command
        await self.call(command, data, parameters)
        # Close the alchemy session
        if session is not None:
            await ru.asyncify(session.close)

    async def run(self):
        self.client = nio.AsyncClient(self.homeserver, self.matrix_id)
        await self.client.login(self.password)
        self._started_timestamp = int(datetime.datetime.now().timestamp() * 1000)
        # matrix-nio type annotations are wrong for asyncclients
        # noinspection PyTypeChecker
        self.client.add_event_callback(self.handle_message, (nio.RoomMessageText,))
        await self.client.sync_forever()



from typing import *
import royalnet
import royalnet.commands as rc
import royalnet.utils as ru
from ..tables import User


class RoyalnetrolesCommand(rc.Command):
    name: str = "royalnetroles"

    description: str = "Display your Royalnet roles."

    syntax: str = ""

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        if name := args.optional(0) is not None:
            user = await User.find(alchemy=self.alchemy, session=data.session, alias=name)
        else:
            user = await data.get_author(error_if_none=True)

        msg = [
            "👤 You currently have these roles:",
            *list(map(lambda r: f"- {r}", user.roles))
        ]

        await data.reply("\n".join(msg))



import royalnet.utils as ru
from royalnet.constellation.api import *


class ApiTokenInfoStar(ApiStar):
    path = "/api/token/info/v1"

    summary = "Get info the current login token."

    tags = ["token"]

    requires_auth = True

    async def api(self, data: ApiData) -> ru.JSON:
        token = await data.token()
        return token.json()



import logging
import sys
import traceback
from typing import *
from royalnet.version import semantic
import functools

try:
    import sentry_sdk
    from sentry_sdk.integrations.aiohttp import AioHttpIntegration
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
    from sentry_sdk.integrations.logging import LoggingIntegration
except ImportError:
    sentry_sdk = None
    AioHttpIntegration = None
    SqlalchemyIntegration = None
    LoggingIntegration = None


log = logging.getLogger(__name__)


def init_sentry(sentry_cfg: Dict[str, Any]):
    if sentry_sdk is None:
        raise ImportError("`sentry` extra is not installed")
    log.debug("Initializing Sentry...")
    release = f"royalnet@{semantic}"
    sentry_sdk.init(sentry_cfg["dsn"],
                    integrations=[AioHttpIntegration(),
                                  SqlalchemyIntegration(),
                                  LoggingIntegration(event_level=None)],
                    release=release)
    log.info(f"Sentry: {release}")


# noinspection PyUnreachableCode
def sentry_exc(exc: Exception,
               level: str = "ERROR"):
    if sentry_sdk is not None:
        with sentry_sdk.configure_scope() as scope:
            scope.set_level(level.lower())
            sentry_sdk.capture_exception(exc)
    level_int: int = logging._nameToLevel[level.upper()]
    log.log(level_int, f"Captured {level.capitalize()}: {exc}")
    # If started in debug mode (without -O), raise the exception, allowing you to see its source
    if __debug__:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback)


def sentry_wrap(level: str = "ERROR"):
    def decorator(func: Callable):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as exc:
                sentry_exc(exc=exc, level=level)
                raise
        return new_func
    return decorator


def sentry_async_wrap(level: str = "ERROR"):
    def decorator(func: Callable):
        @functools.wraps(func)
        async def new_func(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as exc:
                sentry_exc(exc=exc, level=level)
                raise
        return new_func
    return decorator



import asyncio
import datetime


async def sleep_until(dt: datetime.datetime) -> None:
    """Sleep until the specified datetime.

    Warning:
        Accurate only to seconds."""
    now = datetime.datetime.now()
    if now > dt:
        return
    delta = dt - now
    await asyncio.sleep(delta.total_seconds())



from typing import *
import asyncio as aio
from .errors import UnsupportedError
from .configdict import ConfigDict

if TYPE_CHECKING:
    from .event import Event
    from .command import Command
    from ..alchemy import Alchemy
    from ..serf import Serf
    from ..constellation import Constellation


class CommandInterface:
    name: str = NotImplemented
    """The name of the :class:`CommandInterface` that's being implemented.
    
    Examples:
        ``telegram``, ``discord``, ``console``..."""

    prefix: str = NotImplemented
    """The prefix used by commands on the interface.
    
    Examples:
        ``/`` on Telegram, ``!`` on Discord."""

    serf: Optional["Serf"] = None
    """A reference to the :class:`~royalnet.serf.Serf` that is implementing this :class:`CommandInterface`.
    
    Example:
        A reference to a :class:`~royalnet.serf.telegram.TelegramSerf`."""

    constellation: Optional["Constellation"] = None
    """A reference to the Constellation that is implementing this :class:`CommandInterface`.
    
    Example:
        A reference to a :class:`~royalnet.constellation.Constellation`."""

    def __init__(self, config: Dict[str, Any]):
        self.config: ConfigDict[str, Any] = ConfigDict.convert(config)
        """The config section for the pack of the command."""

        # Will be bound after the command/event has been created
        self.command: Optional[Command] = None
        self.event: Optional[Event] = None

    @property
    def alchemy(self) -> "Alchemy":
        """A shortcut for :attr:`.serf.alchemy`."""
        return self.serf.alchemy

    @property
    def table(self) -> "Callable":
        """A shortcut for :func:`.serf.alchemy.get`.

        Raises:
            UnsupportedError: if :attr:`.alchemy` is :const:`None`."""
        if self.alchemy is None:
            raise UnsupportedError("'alchemy' is not enabled on this Royalnet instance")
        return self.alchemy.get

    @property
    def loop(self) -> aio.AbstractEventLoop:
        """A shortcut for :attr:`.serf.loop`."""
        if self.serf:
            return self.serf.loop
        raise UnsupportedError("This command is not being run in a serf.")

    async def call_herald_event(self, destination: str, event_name: str, **kwargs) -> dict:
        """Call an event function on a different :class:`~royalnet.serf.Serf`.

        Example:
            You can run a function on a :class:`~royalnet.serf.discord.DiscordSerf` from a
            :class:`~royalnet.serf.telegram.TelegramSerf`.
        """
        raise UnsupportedError(f"{self.call_herald_event.__name__} is not supported on this platform.")



import logging
from .apierrors import MissingParameterError
from royalnet.backpack.tables.tokens import Token
from royalnet.backpack.tables.users import User
from .apierrors import *
import royalnet.utils as ru

log = logging.getLogger(__name__)


class ApiData(dict):
    def __init__(self, data, star):
        super().__init__(data)
        self.star = star
        self._session = None

    def __missing__(self, key):
        raise MissingParameterError(f"Missing '{key}'")

    async def token(self) -> Token:
        token = await Token.find(self.star.alchemy, self.session, self["token"])
        if token is None:
            raise ForbiddenError("'token' is invalid")
        if token.expired:
            raise ForbiddenError("Login token has expired")
        return token

    async def user(self) -> User:
        return (await self.token()).user

    @property
    def session(self):
        if self._session is None:
            if self.star.alchemy is None:
                raise UnsupportedError("'alchemy' is not enabled on this Royalnet instance")
            log.debug("Creating Session...")
            self._session = self.star.alchemy.Session()
        return self._session

    async def session_commit(self):
        """Asyncronously commit the :attr:`.session` of this object."""
        if self._session:
            log.warning("Session had to be created to be committed")
        # noinspection PyUnresolvedReferences
        log.debug("Committing Session...")
        await ru.asyncify(self.session.commit)

    async def session_close(self):
        """Asyncronously close the :attr:`.session` of this object."""
        if self._session is not None:
            log.debug("Closing Session...")
            await ru.asyncify(self._session.close)



import json
from typing import *
from royalnet.constellation import PageStar
from royalnet.constellation.api import ApiStar
from starlette.requests import Request
from starlette.responses import Response, HTMLResponse
from royalnet.version import semantic


class DocsStar(PageStar):
    path = "/docs"

    async def page(self, request: Request) -> Response:
        paths = {}

        for star in self.constellation.stars:
            if not isinstance(star, ApiStar):
                continue
            paths[star.path] = star.swagger()

        spec = json.dumps({
            "openapi": "3.0.0",
            "info": {
                "description": "Autogenerated Royalnet API documentation",
                "title": "Royalnet",
                "version": f"{semantic}",
            },
            "paths": paths,
            "components": {
                "securitySchemes": {
                    "RoyalnetLoginToken": {
                        "type": "apiKey",
                        "in": "query",
                        "name": "token",
                    }
                }
            }
        })

        return HTMLResponse(
            f"""
            <html lang="en">
                <head>
                    <title>Royalnet Docs</title>
                    <link rel="stylesheet" type="text/css" href="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui.css">
                    <script src="https://unpkg.com/swagger-ui-dist@3/swagger-ui-bundle.js"></script>
                    <script src="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui-standalone-preset.js"></script>
                </head>
                <body>
                    <div id="docs"/>
                    <script>
                        const ui = SwaggerUIBundle({{
                            spec: JSON.parse('{spec}'),
                            dom_id: '#docs',
                            presets: [
                                SwaggerUIBundle.presets.apis,
                                SwaggerUIStandalonePreset
                            ],
                            layout: "StandaloneLayout"
                        }})
                    </script>
                </body>
            </html>
            """
        )



from typing import *
from .commandinterface import CommandInterface
from .commandargs import CommandArgs
from .commanddata import CommandData


class Command:
    name: str = NotImplemented
    """The main name of the command.
    
    Example:
        To be able to call ``/example`` on Telegram, the name should be ``"example"``."""

    aliases: List[str] = []
    """A list of possible aliases for a command.
    
    Example:
        To be able to call ``/e`` as an alias for ``/example``, one should set aliases to ``["e"]``."""

    description: str = NotImplemented
    """A small description of the command, to be displayed when the command is being autocompleted."""

    syntax: str = ""
    """The syntax of the command, to be displayed when a :py:exc:`InvalidInputError` is raised,
     in the format ``(required_arg) [optional_arg]``."""

    def __init__(self, interface: CommandInterface):
        self.interface = interface

    def __str__(self):
        return f"[c]{self.interface.prefix}{self.name}[/c]"

    @property
    def serf(self):
        """A shortcut for :attr:`.interface.serf`."""
        return self.interface.serf

    @property
    def alchemy(self):
        """A shortcut for :attr:`.interface.alchemy`."""
        return self.interface.alchemy

    @property
    def loop(self):
        """A shortcut for :attr:`.interface.loop`."""
        return self.interface.loop

    @property
    def config(self):
        """A shortcut for :attr:`.interface.config`."""
        return self.interface.config

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        raise NotImplementedError()



from typing import *


class Broadcast:
    def __init__(self, handler: str, data: dict, msg_type: Optional[str] = None):
        super().__init__()
        if msg_type is not None:
            assert msg_type == self.__class__.__name__
        self.msg_type = self.__class__.__name__
        self.handler: str = handler
        self.data: dict = data

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.handler == other.handler and self.data == other.data
        return False

    def __repr__(self):
        return f"{self.__class__.__qualname__}(handler={self.handler}, data={self.data})"



class BardError(Exception):
    """Base class for :mod:`bard` errors."""


class YtdlError(BardError):
    """Base class for errors caused by :mod:`youtube_dl`."""


class NotFoundError(YtdlError):
    """The requested resource wasn't found."""


class MultipleFilesError(YtdlError):
    """The resource contains multiple media files."""



import asyncio as aio
from .commandinterface import CommandInterface
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from serf import Serf


class Event:
    """A remote procedure call triggered by a :mod:`royalnet.herald` request."""

    name = NotImplemented
    """The event_name that will trigger this event."""

    def __init__(self, interface: CommandInterface):
        """Bind the event to a :class:`~royalnet.serf.Serf`."""
        self.interface: CommandInterface = interface
        """The :class:`CommandInterface` available to this :class:`Event`."""

    @property
    def serf(self) -> "Serf":
        """A shortcut for :attr:`.interface.serf`."""
        return self.interface.serf

    @property
    def alchemy(self):
        """A shortcut for :attr:`.interface.alchemy`."""
        return self.interface.alchemy

    @property
    def loop(self) -> aio.AbstractEventLoop:
        """A shortcut for :attr:`.interface.loop`."""
        return self.interface.loop

    @property
    def config(self) -> dict:
        """A shortcut for :attr:`.interface.config`."""
        return self.interface.config

    async def run(self, **kwargs):
        raise NotImplementedError()



from royalnet.utils import *
from royalnet.backpack.tables import *
from royalnet.constellation.api import *


class ApiUserFindStar(ApiStar):
    path = "/api/user/find/v1"

    summary = "Find a Royalnet user by one of their aliases."

    tags = ["user"]

    parameters = {
        "alias": "One of the aliases of the user to get."
    }

    async def api(self, data: ApiData) -> dict:
        user = await User.find(self.alchemy, data.session, data["alias"])
        if user is None:
            raise NotFoundError("No such user.")
        return user.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['royalnet',
 'royalnet.alchemy',
 'royalnet.backpack',
 'royalnet.backpack.commands',
 'royalnet.backpack.events',
 'royalnet.backpack.stars',
 'royalnet.backpack.tables',
 'royalnet.backpack.utils',
 'royalnet.bard',
 'royalnet.bard.discord',
 'royalnet.commands',
 'royalnet.constellation',
 'royalnet.constellation.api',
 'royalnet.herald',
 'royalnet.serf',
 'royalnet.serf.discord',
 'royalnet.serf.matrix',
 'royalnet.serf.telegram',
 'royalnet.utils']

package_data = \
{'': ['*']}

install_requires = \
['dateparser>=0.7.2,<0.8.0', 'toml>=0.10.0,<0.11.0']

extras_require = \
{'alchemy_easy': ['sqlalchemy>=1.3.10,<2.0.0',
                  'psycopg2_binary>=2.8.4,<3.0.0',
                  'bcrypt>=3.1.7,<4.0.0'],
 'alchemy_hard': ['sqlalchemy>=1.3.10,<2.0.0',
                  'psycopg2>=2.8.4,<3.0.0',
                  'bcrypt>=3.1.7,<4.0.0'],
 'bard': ['ffmpeg_python>=0.2.0,<0.3.0', 'youtube_dl', 'eyed3>=0.9,<0.10'],
 'coloredlogs': ['coloredlogs>=10.0,<11.0'],
 'constellation': ['starlette>=0.12.13,<0.13.0',
                   'uvicorn>=0.10.7,<0.11.0',
                   'python-multipart>=0.0.5,<0.0.6'],
 'discord': ['discord.py>=1.3.1,<2.0.0', 'pynacl>=1.3.0,<2.0.0'],
 'herald': ['websockets>=8.1,<9.0'],
 'matrix': ['matrix-nio>=0.6,<0.7'],
 'sentry': ['sentry_sdk>=0.13.2,<0.14.0'],
 'telegram': ['python_telegram_bot>=12.2.0,<13.0.0']}

setup_kwargs = {
    'name': 'royalnet',
    'version': '5.8.0',
    'description': 'A multipurpose bot and web framework',
    'long_description': '# `royalnet` [![PyPI](https://img.shields.io/pypi/v/royalnet.svg)](https://pypi.org/project/royalnet/)\n\nA multipurpose bot framework and webserver\n\n## About\n\n`royalnet` is a Python framework that allows you to create interconnected modular chat bots accessible through multiple interfaces (such as Telegram or Discord), and also modular websites that can be connected with the bots.\n\n### Supported bot platforms ("serfs")\n\n- [Telegram](https://core.telegram.org/bots)\n- [Discord](https://discordapp.com/developers/docs/)\n- [Matrix](https://matrix.org/) (no E2E support yet)\n\n## Installing\n\nTo install `royalnet`, run:\n```\npip install royalnet\n```\n\nTo install a specific module, run:\n```\npip install royalnet[MODULENAME]\n```\n\nTo install all `royalnet` modules, run:\n```\npip install royalnet[telegram,discord,matrix,alchemy_easy,bard,constellation,sentry,herald,coloredlogs]\n```\n\n## Documentation\n\n`royalnet`\'s documentation is available [here](https://gh.steffo.eu/royalnet/html).\n\n## Developing `royalnet`\n\nTo develop `royalnet`, you need to have [Poetry](https://poetry.eustace.io/) installed on your PC.\n\nAfter you\'ve installed Poetry, clone the git repo with the command:\n\n```\ngit clone https://github.com/Steffo99/royalnet\n```\n\nThen enter the new directory:\n\n```\ncd royalnet\n```\n\nAnd finally install all dependencies and the package:\n\n```\npoetry install -E telegram -E discord -E matrix -E alchemy_easy -E bard -E constellation -E sentry -E herald -E coloredlogs\n```\n',
    'author': 'Stefano Pigozzi',
    'author_email': 'ste.pigozzi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Steffo99/royalnet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Imports go here!
from .users import User
from .telegram import Telegram
from .discord import Discord
from .matrix import Matrix
from .aliases import Alias
from .tokens import Token
from .roles import Role

# Enter the tables of your Pack here!
available_tables = {
    User,
    Telegram,
    Discord,
    Matrix,
    Alias,
    Token,
    Role,
}

# Don't change this, it should automatically generate __all__
__all__ = [table.__name__ for table in available_tables]



import os
import logging
import re
import eyed3
from contextlib import asynccontextmanager
from typing import *
from royalnet.utils import *
from asyncio import AbstractEventLoop, get_event_loop
from .ytdlinfo import YtdlInfo
from .errors import NotFoundError, MultipleFilesError
from youtube_dl import YoutubeDL


log = logging.getLogger(__name__)


class YtdlFile:
    """A representation of a file download with `youtube_dl <https://ytdl-org.github.io/youtube-dl/index.html>`_."""

    default_ytdl_args = {
        "quiet": not __debug__,  # Do not print messages to stdout.
        "noplaylist": True,  # Download single video instead of a playlist if in doubt.
        "no_warnings": not __debug__,  # Do not print out anything for warnings.
        "outtmpl": "./downloads/%(epoch)s-%(title)s-%(id)s.%(ext)s",  # Use the default outtmpl.
        "ignoreerrors": True  # Ignore unavailable videos
    }

    def __init__(self,
                 url: str,
                 info: Optional[YtdlInfo] = None,
                 filename: Optional[str] = None,
                 ytdl_args: Optional[Dict[str, Any]] = None,
                 loop: Optional[AbstractEventLoop] = None):
        """Create a :class:`YtdlFile` instance.

        Warning:
            Please avoid using directly :meth:`.__init__`, use :meth:`.from_url` instead!"""
        self.url: str = url
        self.info: Optional[YtdlInfo] = info
        self.filename: Optional[str] = filename
        self.ytdl_args: Dict[str, Any] = {**self.default_ytdl_args, **ytdl_args}
        self.lock: MultiLock = MultiLock()
        if not loop:
            loop = get_event_loop()
        self._loop = loop

    def __repr__(self):
        if not self.has_info:
            return f"<{self.__class__.__qualname__} without info>"
        elif not self.is_downloaded:
            return f"<{self.__class__.__qualname__} not downloaded>"
        else:
            return f"<{self.__class__.__qualname__} at '{self.filename}'>"

    @property
    def has_info(self) -> bool:
        """Does the :class:`YtdlFile` have info available?"""
        return self.info is not None

    async def retrieve_info(self) -> None:
        """Retrieve info about the :class:`YtdlFile` through :class:`YoutubeDL`."""
        if not self.has_info:
            infos = await asyncify(YtdlInfo.from_url, self.url, loop=self._loop, **self.ytdl_args)
            if len(infos) == 0:
                raise NotFoundError()
            elif len(infos) > 1:
                raise MultipleFilesError()
            self.info = infos[0]

    @property
    def is_downloaded(self) -> bool:
        """Has the file been downloaded yet?"""
        return self.filename is not None

    async def download_file(self) -> None:
        """Download the file."""
        def download():
            """Download function block to be asyncified."""
            with YoutubeDL(self.ytdl_args) as ytdl:
                filename = ytdl.prepare_filename(self.info.__dict__)
            with YoutubeDL({**self.ytdl_args, "outtmpl": filename}) as ytdl:
                ytdl.download([self.info.webpage_url])
            # "WARNING: Requested formats are incompatible for merge and will be merged into mkv."
            if not os.path.exists(filename):
                filename = re.sub(r"\.[^.]+$", ".mkv", filename)
            self.filename = filename

        await self.retrieve_info()
        if not self.is_downloaded:
            async with self.lock.exclusive():
                log.debug(f"Downloading with youtube-dl: {self}")
                await asyncify(download, loop=self._loop)
                if self.info.extractor == "generic":
                    log.debug(f"Generic extractor detected, updating info from the downloaded file: {self}")
                    self.set_ytdlinfo_from_id3_tags()

    def set_ytdlinfo_from_id3_tags(self):
        tag_file = eyed3.load(self.filename)
        if not tag_file:
            log.debug(f"No ID3 tags found: {self}")
            return
        tag: eyed3.core.Tag = tag_file.tag
        if tag.title:
            log.debug(f"Found title: {self}")
            self.info.title = tag.title
        if tag.album:
            log.debug(f"Found album: {self}")
            self.info.album = tag.album
        if tag.artist:
            log.debug(f"Found artist: {self}")
            self.info.artist = tag.artist

    @asynccontextmanager
    async def aopen(self):
        """Open the downloaded file as an async context manager (and download it if it isn't available yet).

        Example:
            You can use the async context manager like this: ::

                async with ytdlfile.aopen() as file:
                    b: bytes = file.read()

        """
        await self.download_file()
        async with self.lock.normal():
            log.debug(f"File opened: {self.filename}")
            with open(self.filename, "rb") as file:
                yield file
            log.debug(f"File closed: {self.filename}")

    async def delete_asap(self):
        """As soon as nothing is using the file, delete it."""
        log.debug(f"Trying to delete: {self.filename}")
        if self.filename is not None:
            async with self.lock.exclusive():
                os.remove(self.filename)
                log.debug(f"Deleted: {self.filename}")
                self.filename = None

    @classmethod
    async def from_url(cls, url: str, **ytdl_args) -> List["YtdlFile"]:
        """Create a :class:`list` of :class:`YtdlFile` from a URL."""
        infos = await YtdlInfo.from_url(url, **ytdl_args)
        files = []
        for info in infos:
            file = YtdlFile(url=info.webpage_url, info=info, ytdl_args=ytdl_args)
            files.append(file)
        return files
